# Synthesiser for Composition "First Steps"
# Copyright (C) 2010 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# This program contains the specifications for a work of art---a
# musical composition called "First Steps". First Steps by Joshua D
# Bartlett is licensed under a Creative Commons Attribution-Share
# Alike 2.5 Australia License. Please attribute any use of the work
# to Joshua D Bartlett, and where possible, link to:
#   http://sqizit.blogetery.com/projects/impossible-music
# The full text of the license should be included with this file.
# For information on the license, please see the following website:
#   http://creativecommons.org/licenses/by-sa/2.5/au/
#
from __future__ import division
import sys

from wavemaker import (Composition, NoteGroup, PureNoteGenerator, Bar,
        play_sound, Progression, SmoothVoice, HarshVoice, main)

DEBUG = False
SLOW = False

def one_alberti_bass(voice, key, duration):
    return voice.notes(
        (4/8*key, duration),
        (6/8*key, duration),
        (5/8*key, duration),
        (6/8*key, duration),
    )

def alberti_bass(voice, key=1, duration=1/2, repetitions=2):
    result = NoteGroup()
    for i in xrange(repetitions):
        result.extend(one_alberti_bass(voice, key, duration))
    return result

if DEBUG:
    sample_rate = 11025
else:
    sample_rate = 44100

if SLOW:
    bpm = 40
else:
    bpm = 80

def composition1():
    c = Composition(sample_rate=sample_rate, beats_per_minute=bpm)
    treb = HarshVoice(c, 'treb', amplitude=0.15)
    bass = HarshVoice(c, 'bass', amplitude=0.1)

    root_bass = alberti_bass(voice=bass, key=1)
    dominant_bass = alberti_bass(voice=bass, key=3/4)
    submed_bass = alberti_bass(voice=bass, key=4/5)
    median_bass = alberti_bass(voice=bass, key=3/5)
    
    c.add(
        Bar(
            root_bass,
            treb.notes(
                ( 8/8, 1),
                (10/8, 1/2),
                ( 9/8, 1/2),
                (12/8, 1/2),
                (14/8, 1/4),
                (13/8, 1/4),
                (12/8, 1/2),
                (3/2*9/8, 1/2),
            ),
        ),
        Bar(
            Progression(
                alberti_bass(voice=bass, key=3/4, repetitions=1),
                bass.notes(
                    (3/4*  4/8,   1/2),
                    (3/4*  6/8,   1/2),
                    (3/4* 57/92,  1/2),
                    (3/4*190/253, 1/2),
                )
            ),
            treb.notes(
                ( 3/2* 10/8, 1),
                ( 3/2*  8/8, 1/2),
                ( 3/2*  7/8, 1/2),
                ( 3/2*  7/8, 1),
                ( 3/2*  19/23, 1),
            ),
        ),
        Bar(
            submed_bass,
            treb.notes(
                ( 4/5* 16/8, 1),
                ( 4/5* 15/8, 1/2),
                ( 4/5* 14/8, 1/2),
                ( 4/5* 14/8, 1),
                ( 4/5* 15/8, 1),
            ),
        ),
        Bar(
            Progression(
                alberti_bass(voice=bass, key=3/5, repetitions=1),
                alberti_bass(voice=bass, key=9/10, repetitions=1),
            ),
            treb.notes(
                (3/5* 10/4, 3/4),
                (3/5* 11/4, 1/4),
                (3/5* 12/4, 1/2),
                (3/5* 10/4, 1/2),
                (3/5* 9/4, 2),
            ),
        ),
        Bar(
            bass.notes(
                (7/20,        1/2),
                (7/20*  3/2,  1/2),
                (7/20* 19/15, 1/2),
                (7/20*  3/2,  1/2),
                (7/20,        1/2),
                (7/20*  3/2,  1/2),
                (7/20* 19/15, 1/2),
                (7/20*  3/2,  1/2),
            ),
            treb.notes(
                (7/5, 1),
                (7/5* 19/20, 1/2),
                (7/5*  9/10, 1/2),
                (7/5*  6/7, 1/2),
                (7/5*(19/20), 1/4),
                (7/5* 11/13, 1/4),
                (7/5* 3/4, 1),
            ),
        ),
        Bar(
            Progression(
                alberti_bass(voice=bass, key=3/5, repetitions=1),
                alberti_bass(voice=bass, key=9/10, repetitions=1),
            ),
            treb.notes(
                (12/10*  3/2, 1),
                (12/10*  4/3, 1/2),
                (12/10*  5/4, 1/2),
                ( 9/10*  5/3, 1),
                ( 9/10*  3/2, 1),
            ),
        ),
        Bar(
            dominant_bass,
            treb.notes(
                (3/2* 12/8, 1),
                (3/2* 12/8, 1/2),
                (3/2* 1.3, 1/2),
                (3/2* 10/8, 1/2),
                (3/2* 10/8, 1/4),
                (3/2*  9/8, 1/4),
                (3/2*  8/8, 1/2),
                (3/2*  9/8, 1/2),
            ),
        ),
        Bar(
            Progression(
                alberti_bass(voice=bass, key=1, repetitions=1),
                bass.notes(
                    (1/2, 1/2),
                    (3/4, 1/2),
                    (1/2, 1),
                ),
            ),
            treb.notes(
                (3/2, 1/2),
                (17/12, 1/4),
                (16/12, 1/4),
                (5/4, 1/2),
                (9/8, 1/2),
                (9/8, 1),
                (1, 1),
            ),
        ),
    )

    return c

main(composition1)
