# WaveMaker Synthesiser Software
# Copyright (C) 2010 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
from __future__ import division

import inspect
from math import sin, cos, pi, asin
import os
import StringIO
import struct
import sys
import time
import wave

class Composition(object):
    def __init__(self, sample_rate=44100, sample_width=2, num_channels=1,
            fill_fraction=0.95, root_pitch=220, beats_per_minute=80,
            file=None, initial_silence=0.5, final_silence=None,
            default_channel=0, amplitude=0.1):
        self.sample_rate = sample_rate
        self.sample_width = sample_width
        self.num_channels = num_channels
        self.fill_fraction = fill_fraction
        self.root_pitch = root_pitch
        self.beats_per_minute = beats_per_minute
        self.file = file
        self.initial_silence = initial_silence
        self.final_silence = final_silence
        self.default_channel = default_channel
        self.amplitude = amplitude

        self.validate()
        self.initialise()

    def initialise(self):
        self.notegroup = NoteGroup()

    def validate(self):
        self.validate_sample_width()
        self.validate_file()
        self.validate_silence()
        self.validate_channels()
        self.validate_fill_fraction()
        self.validate_amplitude()

    def validate_sample_width(self):
        sw = self.sample_width
        if sw == 4:
            self.pack_spec = '<L'
        elif sw == 2:
            self.pack_spec = '<h'
        else:
            raise ValueError('sample_width must be 2 or 4 for now')
        
    def validate_file(self):
        if self.file is None:
            self.file = StringIO.StringIO()

    def validate_silence(self):
        if self.final_silence is None:
            self.final_silence = self.initial_silence

    def validate_channels(self):
        if self.num_channels < 1:
            raise ValueError('num_channels must be 1 or greater')
        if self.default_channel < 0:
            raise ValueError('default_channel must be 0 or greater')
        if self.default_channel >= self.num_channels:
            raise ValueError('default_channel must be less than num_channels')

    def validate_fill_fraction(self):
        if self.fill_fraction <= 0:
            raise ValueError('fill_fraction must be greater than 0')
        elif self.fill_fraction > 1:
            raise ValueError('fill_fraction may not be greater than 1')

    def validate_amplitude(self):
        if self.amplitude > 1 or self.amplitude < 0:
            raise ValueError('amplitude must be from 0 to 1')

    def play(self, wait_for_completion=True, verbose=True):
        '''
        Plays using pygame.mixer.
        '''
        file = StringIO.StringIO()
        if verbose:
            print 'Rendering wave data...'
        self.render(file)

        file.seek(0)
        if verbose:
            print 'Playing wave sound.'
        play_sound(file, self.sample_rate, self.sample_width,
                self.num_channels, wait_for_completion)

    def render(self, file=None):
        if file is None:
            file = self.file
        w = wave.open(file, 'w')
        w.setnchannels(self.num_channels)
        w.setsampwidth(self.sample_width)
        w.setframerate(self.sample_rate)
        self.render_body(w)
        w.close()

    def render_body(self, w):
        duration = (self.notegroup.duration + self.initial_silence + 
                self.final_silence)
        iterators = set()
        for voice, notes in self.notegroup.notes.iteritems():
            notes = [Rest(self.initial_silence)] + notes
            iterators.add(voice.render(self.sample_rate, notes))
            
        for i in xrange(int(duration * self.sample_rate + 0.5)):
            value = 0
            for it in list(iterators):
                try:
                    value += it.next()
                except StopIteration:
                    iterators.remove(it)
            w.writeframesraw(self.encode_data(value))

    def encode_data(self, value):
        n = max(min(value, 1), -1)
        x = int(n * (1 << (7*self.sample_width)) + 0.5)
        return struct.pack(self.pack_spec, x)

    def add_voice(self, voice):
        pass

    def add(self, *notegroups):
        for notegroup in notegroups:
            self.notegroup.extend(notegroup)

class Voice(object):
    def __init__(self, composition, name=None, channel=None,
            fill_fraction=None,
            root_pitch=None, beats_per_minute=None, amplitude=None):
        self.composition = composition
        self.name = name
        self.channel = channel
        self.fill_fraction = fill_fraction
        self.root_pitch = root_pitch
        self.beats_per_minute = beats_per_minute
        self.amplitude = amplitude

        self.fill_defaults()
        self.validate()
        self.initialise()

    def __repr__(self):
        if self.name is not None:
            return self.name
        return object.__repr__(self)

    def initialise(self):
        self.composition.add_voice(self)

    def fill_defaults(self):
        if self.channel is None:
            self.channel = self.composition.default_channel
        if self.fill_fraction is None:
            self.fill_fraction = self.composition.fill_fraction
        if self.root_pitch is None:
            self.root_pitch = self.composition.root_pitch
        if self.beats_per_minute is None:
            self.beats_per_minute = self.composition.beats_per_minute
        if self.amplitude is None:
            self.amplitude = self.composition.amplitude

    def validate(self):
        self.validate_fill_fraction()
        self.validate_amplitude()

    def validate_fill_fraction(self):
        if self.fill_fraction <= 0:
            raise ValueError('fill_fraction must be greater than 0')
        elif self.fill_fraction > 1:
            raise ValueError('fill_fraction may not be greater than 1')
            
    def validate_amplitude(self):
        if self.amplitude > 1 or self.amplitude < 0:
            raise ValueError('amplitude must be from 0 to 1')

    def notes(self, *notespecs):
        result = NoteGroup()
        for note in notespecs:
            if not isinstance(note, tuple):
                raise TypeError('each argument to notes() must be a tuple')
            if len(note) != 2:
                raise TypeError('each argument to notes() must have two elements')
            pitch, duration = note
            full_duration = duration * 60 / self.beats_per_minute
            if pitch is None:
                result.append(self, Rest(full_duration))
            else:
                frequency = pitch * self.root_pitch
                on_duration = full_duration * self.fill_fraction
                off_duration = full_duration - on_duration
                result.append(self, Note(frequency, on_duration))
                if off_duration > 0:
                    result.append(self, Rest(off_duration))

        return result

    def render(self, sample_rate, notes):
        raise NotImplementedError('Voice.render')

class PureNoteGenerator(Voice):
    def render(self, sample_rate, notes):
        notes = list(notes)
        surplus = 0
        while len(notes) > 0:
            note = notes.pop(0)
            ticks_to_fill = surplus + int(note.duration * sample_rate + 0.5)
            if ticks_to_fill <= 0:
                continue

            if note.frequency is None:
                for i in xrange(ticks_to_fill):
                    yield 0
                surplus = 0
            else:
                omega = 2 * pi * note.frequency
                num_cycles = int(note.frequency*ticks_to_fill/sample_rate+0.5)
                actual_duration = num_cycles / note.frequency
                ticks_filled = int(sample_rate * actual_duration + 0.5)
                for i in xrange(ticks_filled):
                    t = i / sample_rate
                    yield self.amplitude * sin(omega * t)
                surplus = ticks_to_fill - ticks_filled

class HarmonicGenerator(Voice):
    def __init__(self, composition, intensity_profile, *args, **kwargs):
        super(HarmonicGenerator, self).__init__(composition, *args, **kwargs)
        if 'amplitude' in kwargs:
            del kwargs['amplitude']
        self.pure_generators = []
        for intensity in intensity_profile:
            intensity *= self.amplitude
            gen = PureNoteGenerator(composition, *args, amplitude=intensity, **kwargs)
            self.pure_generators.append(gen)

    def render(self, sample_rate, notes):
        iterators = set()
        for i in xrange(0, len(self.pure_generators)):
            gen = self.pure_generators[i]
            gen_notes = []
            for note in notes:
                n = note.duplicate()
                if n.frequency is not None:
                    n.frequency *= i + 1
                gen_notes.append(n)
            iterators.add(gen.render(sample_rate, gen_notes))
        while len(iterators) > 0:
            value = 0
            for i in list(iterators):
                try:
                    value += i.next()
                except StopIteration:
                    iterators.remove(i)
            yield value
            
def SmoothVoice(composition, *args, **kwargs):
    return HarmonicGenerator(composition,
            [0.5, 0.35, 0.25, 0.175, 0.125], *args, **kwargs)

def HarshVoice(composition, *args, **kwargs):
    return HarmonicGenerator(composition,
            [0.25, 0.35, 0.5, 0.25, 0.125], *args, **kwargs)

def Rest(duration):
    return Note(None, duration)
class Note(object):
    def __init__(self, frequency, duration):
        self.frequency = frequency
        self.duration = duration
    def __repr__(self):
        if self.frequency is None:
            return 'Rest(%r)' % (self.duration,)
        return 'Note(%r, %r)' % (self.frequency, self.duration)
    def duplicate(self):
        return Note(self.frequency, self.duration)

class NoteGroup(object):
    def __init__(self):
        self.duration = 0
        self.notes = {}     # voice -> [notes]
        self.times = {}     # voice -> time
        self.barlines = []  # List of times.

    def append_barline(self):
        if self.duration == 0:
            raise ValueError('Cannot append barline to empty NoteGroup')
        if (len(self.barlines) > 0 and self.barlines[-1]
                == self.duration):
            raise ValueError('NoteGroup already ends with a barline')
        self.barlines.append(self.duration)

    def append(self, voice, note, time=None):
        if not isinstance(note, Note):
            raise TypeError('note must be Note or Rest object')
        if time is None:
            time = self.duration
        if voice not in self.notes:
            self.notes[voice] = []
            self.times[voice] = 0
        last_time = self.times[voice]
        if time < last_time:
            raise ValueError('voice is already playing note at that time')
        if time > last_time:
            self.notes[voice].append(Rest(time - last_time))
        self.notes[voice].append(note)
        self.times[voice] = d = time + note.duration
        self.duration = max(self.duration, d)

    def extend(self, other):
        if not isinstance(other, NoteGroup):
            raise TypeError('argument to extend must be a NoteGroup')
        start_time = self.duration
        for voice, notes in other.notes.iteritems():
            if voice not in self.notes:
                self.notes[voice] = []
                self.times[voice] = 0
            last_time = self.times[voice]
            if start_time > last_time:
                self.notes[voice].append(Rest(start_time - last_time))
            self.notes[voice].extend(notes)
            self.times[voice] = start_time + other.times[voice]
        for barline in other.barlines:
            self.barlines.append(barline + start_time)
        self.duration += other.duration

    def extend_sideways(self, other, start_time=0):
        '''
        Adds the notes in the given note group to this one, starting at the
        given time.
        '''
        if start_time < 0:
            raise ValueError('start_time must be 0 or greater')
        for voice in self.notes:
            if voice in other.notes:
                raise ValueError('voice cannot play two things at '
                        'once: %r' % (voice,))
        for voice in other.notes:
            if voice in self.notes:
                raise ValueError('voice cannot play two things at '
                        'once: %r' % (voice,))
        self.barlines = self._check_barlines(other, start_time)
        for voice, notes in other.notes.iteritems():
            self.notes[voice] = list(notes)
            if start_time > 0:
                self.notes[voice].insert(0, Rest(start_time))
            self.times[voice] = t = other.times[voice] + start_time
            if t > self.duration:
                self.duration = t

    def _check_barlines(self, other, start_time):
        '''
        Checks that the barlines match up between the note groups.
        Returns the new set of barlines.
        '''
        s_barlines = iter(self.barlines)
        o_barlines = iter(other.barlines)
        s_barline = 0
        # Find self's first barline in the matching section.
        while s_barline <= start_time:
            try:
                s_barline = s_barlines.next()
            except StopIteration:
                try:
                    o_barline = o_barlines.next() + start_time
                except StopIteration:
                    return []
                if o_barline > self.duration:
                    return [x + start_time for x in other.barlines]
                raise ValueError('barlines do not match at %f seconds' %
                        (o_barline,))

        result = list(self.barlines)
        while True:
            # Find other's next barline in this section.
            try:
                o_barline = o_barlines.next() + start_time
            except StopIteration:
                if s_barline > other.duration:
                    return self.barlines
                raise ValueError('barlines do not match at %f seconds' %
                        (s_barline,))

            if s_barline != o_barline:
                raise ValueError('barlines do not match at %f seconds' %
                        (min(s_barline, o_barline)))
            result.append(s_barline)

            # Get self's next barline.
            try:
                s_barline = s_barlines.next()
            except StopIteration:
                try:
                    o_barline = o_barlines.next() + start_time
                except StopIteration:
                    return result
                if o_barline > self.duration:
                    while True:
                        result.append(o_barline)
                        try:
                            o_barline = o_barlines.next() + start_time
                        except StopIteration:
                            return result
                raise ValueError('barlines do not match at %f seconds' %
                        (o_barline,))

        
def Simultaneous(*notegroups):
    result = NoteGroup()
    for ng in notegroups:
        result.extend_sideways(ng)
    return result

def Progression(*notegroups):
    result = NoteGroup()
    for ng in notegroups:
        result.extend(ng)
    return result
    
def Silence(duration):
    result = NoteGroup()
    result.duration = duration
    return result

def Bar(*notegroups):
    result = Simultaneous(*notegroups)
    prev_end_time = None
    prev_voice = None
    for voice, end_time in result.times.iteritems():
        if prev_end_time is None:
            prev_end_time = end_time
            prev_voice = voice
            continue
        if prev_end_time != end_time:
            raise ValueError('voices play for different length of time in '
                    'bar: %r (%.2fs), %r (%.2fs)' % (prev_voice, prev_end_time,
                    voice, end_time))
    result.append_barline()
    return result

def play_sound(file, sample_rate=44100, sample_width=2, num_channels=1,
        wait_for_completion=True):
    import pygame.mixer
    pygame.mixer.init(
        sample_rate,
        8*sample_width,
        num_channels,
    )
    s = pygame.mixer.Sound(file)
    s.play()

    if wait_for_completion:
        while pygame.mixer.get_busy():
            time.sleep(0.5)
    pygame.mixer.quit()
    
def alberti_bass(voice, key=1, duration=1):
    return voice.notes(
        (4/4*key, duration),
        (6/4*key, duration),
        (5/4*key, duration),
        (6/4*key, duration),
    )

def demo():
    c = Composition()
    treb = PureNoteGenerator(c)
    bass = PureNoteGenerator(c)

    c.add(Bar(
        alberti_bass(voice=bass, key=1, duration=1/2)
    ))
    c.add(Bar(
        alberti_bass(voice=bass, key=1, duration=1/2)
    ))

    c.play()

def main(composition):
    callerframe = inspect.currentframe().f_back
    if callerframe.f_globals['__name__'] != '__main__':
        return

    pyfile = callerframe.f_code.co_filename
    head, tail = os.path.splitext(pyfile)
    outfile = head + '.wav'
    if (not os.path.isfile(outfile) or os.path.getmtime(pyfile) >=
            os.path.getmtime(outfile) or '--force' in sys.argv):
        print 'Saving composition: %s' % (outfile,)
        composition().render(file=outfile)
    else:
        print 'Source file has not changed.'
    print 'Playing composition.'
    try:
        play_sound(outfile)
    except ImportError:
        print 'Unable to play composition without pygame installed.'

if __name__ == '__main__':
    demo()
